//! Static in-place query construction.
//!
//! Potential interfaces:
//! - `microrm::select().by(KVStore::Key, &key).result(qi)`
//! - `qi.select().by(KVStore::Key, &key).result()`

use crate::{entity::EntityColumn, model::Modelable, Entity, Error};
use std::{marker::PhantomData, collections::HashMap, hash::Hasher};

use super::QueryInterface;


#[derive(Debug,PartialEq,Eq,PartialOrd,Ord,Hash)]
pub enum QueryPart {
    Root,
    Where
}

pub type DerivedQuery = HashMap<QueryPart, Vec<String>>;

trait StaticVersion {
    type Is: 'static;

    fn type_id() -> std::any::TypeId {
        std::any::TypeId::of::<Self::Is>()
    }
}

/// Any query component
trait QueryComponent: StaticVersion {
    fn derive(&self) -> DerivedQuery {
        DerivedQuery::new()
    }

    /// returns the next index to use for binding
    fn bind(&self, stmt: &mut sqlite::Statement<'_>) -> Result<usize, Error>;
}

/// Any query that can be completed/executed
trait Resolvable<'r, 'q, T: Entity>: QueryComponent {

    fn qi(&self) -> &'r QueryInterface<'q>;

    fn no_result(self) where Self: Sized { todo!() }
    fn result(self) -> Option<T> where Self: Sized { todo!() }
    fn results(self) -> Vec<T> where Self: Sized {

        // std::any::TypeId::of::<T>();

        println!("derivation: {:?}", self.derive());

        todo!()
    }
}

/// Any query that can have a WHERE clause attached to it
trait Filterable<'r, 'q>: Resolvable<'r, 'q, Self::Table> {
    type Table: Entity;

    fn by<C: EntityColumn<Entity = Self::Table>, G: Modelable + ?Sized>(self, col: C, given: &'r G) -> Filter<'r, 'q, Self, C, G> where Self: Sized {
        Filter { wrap: self, col, given, _ghost: PhantomData }
    }
}

struct Select<'r, 'q, T: Entity> {
    qi: &'r QueryInterface<'q>,
    _ghost: PhantomData<T>,
}

impl <'r, 'q, T: Entity> StaticVersion for Select<'r, 'q, T> {
    type Is = Select<'static, 'static, T>;
}

impl<'r, 'q, T: Entity> QueryComponent for Select<'r, 'q, T> {
    fn derive(&self) -> DerivedQuery {
        HashMap::from([
            (QueryPart::Root, vec![format!("SELECT * FROM {}", T::table_name())])
        ])
    }

    // next binding point is the first
    fn bind(&self, _stmt: &mut sqlite::Statement<'_>) -> Result<usize, Error> { Ok(1) }
}

impl<'r, 'q, T: Entity> Filterable<'r, 'q> for Select<'r, 'q, T> {
    type Table = T;
}

impl<'r, 'q, T: Entity> Resolvable<'r, 'q, T> for Select<'r, 'q, T> {
    fn qi(&self) -> &'r QueryInterface<'q> { self.qi }
}

/// A concrete WHERE clause
struct Filter<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> {
    wrap: F,
    col: C,
    given: &'r G,
    _ghost: PhantomData<&'q ()>
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> StaticVersion for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static> {
    type Is = Filter<'static, 'static, <F as StaticVersion>::Is, C, u64>;
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> QueryComponent for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static> {
    fn derive(&self) -> DerivedQuery {
        let mut parent = self.wrap.derive();

        parent.entry(QueryPart::Where).or_insert_with(|| Vec::new()).push(
            format!("{} = ?", self.col.name()));

        // println!("derivation: {:?}", self.derive());

        parent
    }

    fn bind(&self, stmt: &mut sqlite::Statement<'_>) -> Result<usize, crate::Error> {
        let next_index = self.wrap.bind(stmt)?;

        self.given.bind_to(stmt, next_index)?;

        Ok(next_index + 1)
    }
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> Resolvable<'r, 'q, C::Entity> for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static> {
    fn qi(&self) -> &'r QueryInterface<'q> { self.wrap.qi() }
}

impl<'r, 'q, F: Filterable<'r, 'q>, C: EntityColumn, G: Modelable + ?Sized> Filterable<'r, 'q> for Filter<'r, 'q, F, C, G> where <F as StaticVersion>::Is: Filterable<'static, 'static> {
    type Table = C::Entity;
}

#[cfg(test)]
mod test_build {
    use microrm_macros::Entity;
    use serde::{Deserialize, Serialize};

    use crate::QueryInterface;

    #[derive(Entity, Serialize, Deserialize)]
    #[microrm_internal]
    pub struct KVStore {
        key: String,
        value: String,
    }

    fn simple_test() {
        use super::*;
        let db = crate::DB::new_in_memory(crate::Schema::new().entity::<KVStore>()).unwrap();
        let qi = db.query_interface();

        let select = Select::<KVStore> { qi: &qi, _ghost: PhantomData };
        // drop(select);
        let filter = select.by(KVStore::Key, "abc");
        filter.results();
    }

    #[test]
    fn simple_get() {
        simple_test()
    }
}
